from svf_package.grid.svfgrid import SVFGrid
from tests import data

if __name__ == '__main__':

    inputs = ["x1","x2","y1"]
    outputs = ["y1","y2"]
    d = 2

    grid = SVFGrid(data, inputs, outputs, d)
    grid.create_grid()

    print(grid.data_grid)


from itertools import product
from numpy import arange
from pandas import DataFrame
from svf_package.grid.grid import GRID


class SVFGrid(GRID):
    """
        Clase generadora de un grid SVF. Sirve tanto para SVF como SSVF
    """

    def __init__(self, data, inputs, outputs, d):
        """
            Constructor de la clase SVFGrid
        Args:
            data (pandas.DataFrame): conjunto de datos sobre los que se construye el grid
            inputs (list): listado de inputs
            d (list): número de particiones en las que se divide el grid
        """
        super().__init__(data, inputs, outputs, d)
        self.df_grid = None

    def create_grid(self):
        """
            Función que crea un grid en base a unos datos e hiperparámetro d
        """
        self.df_grid = DataFrame(columns=["id_cell", "value", "phi"])
        x = self.data.filter(self.inputs)
        # Numero de columnas x
        n_dim = len(x.columns)
        # Lista de listas de knot
        knot_list = list()
        # Lista de indices (posiciones) para crear el vector de subind
        knot_index = list()
        for col in range(0, n_dim):
            # knots de la dimension col
            knot = list()
            knot_max = x.iloc[:, col].max()
            knot_min = x.iloc[:, col].min()
            amplitud = (knot_max - knot_min) / self.d
            for i in range(0, self.d + 1):
                knot_i = knot_min + i * amplitud
                knot.append(knot_i)

            knot_list.append(knot)
            knot_index.append(arange(0, len(knot)))
        self.df_grid["id_cell"] = list(product(*knot_index))
        self.df_grid["value"] = list(product(*knot_list))
        self.knot_list = knot_list
        self.calculate_df_grid()
        self.calculate_data_grid()

    def calculate_dmu_phi(self, cell):
        """
            Función que calcula el valor de la transformación (phi) de una observación en el grid.
        Args:
            cell (list): Posición de la observación en el grid

        Returns:
            list: Vector de 1 0 con la transformación del vector en base al grid
        """
        phi = []
        phi_list = []
        n_dim = len(cell)
        value = 0
        for i in range(0, len(self.df_grid)):
            for j in range(0, n_dim):
                if cell[j] >= self.df_grid["id_cell"][i][j]:
                    value = 1
                else:
                    value = 0
                    break
            phi.append(value)
        for i in range(len(self.outputs)):
            phi_list.append(phi)
        return phi_list

    def calculate_df_grid(self):
        """Método para añadir al dataframe grid el valor de la transformada de cada observación
        """
        x = self.df_grid["value"]
        x_list = x.values.tolist()
        phi_list = list()
        c_cells = list()
        for x in x_list:
            p = self.search_dmu(x)
            phi = self.calculate_dmu_phi(p)
            phi_list.append(phi)
        for index, cell in self.df_grid.iterrows():
            c_cell = search_contiguous_cell(cell['id_cell'])
            c_cells.append(c_cell)
        self.df_grid["phi"] = phi_list
        self.df_grid["c_cells"] = c_cells

    def calculate_data_grid(self):
        """Método para añadir al dataframe grid el valor de la transformada de cada observación
        """
        self.data_grid = self.data.copy()
        x = self.data_grid.filter(self.inputs)
        x_list = x.values.tolist()
        phi_list = list()
        c_cells = list()
        for x in x_list:
            p = self.search_dmu(x)
            phi = self.calculate_dmu_phi(p)
            phi_list.append(phi)
            c_cell = search_contiguous_cell(p)
            c_cells.append(c_cell)
        self.data_grid["phi"] = phi_list
        self.data_grid["c_cells"] = c_cells


def search_contiguous_cell(cell):
    con_c_list = list()
    cell = list(cell)
    for dim in range(len(cell)):
        value = (cell[dim]) - 1
        con_cell = cell.copy()
        if value >= 0:
            con_cell[dim] = value
            con_c_list.append(tuple(con_cell))
    return con_c_list



class SVFDualSolution:
    """Clase solución de SVF
    """
    def __init__(self, gamma, alpha, delta, w):

        """Constructor de la clase solución SVF en los forma dual

        Args:
            gamma (list): Valores de la variable gamma del modelo ya solucionado
            alpha (list): Valores de la variable alpha del modelo ya solucionado
            delta (list): Valores de la variable delta del modelo ya solucionado
            w (list): Valores de los pesos w del modelo ya solucionado
        """

        self.gamma = gamma
        self.alpha = alpha
        self.delta = delta
        self.w = w



from docplex.mp.model import Model
from svf_package.grid.svf_splines_grid import SVFSplinesGrid
from svf_package.methods.svf import SVF
from svf_package.solution.svf_solution import SVFPrimalSolution


class SVFSplines(SVF):
    """Clase del modelo SVF Splines
    """

    def __init__(self, method, inputs, outputs, data, C, eps, d):
        """Constructor de la clase SVF Splines

        Args:
            method (string): Método SVF que se quiere utilizar
            inputs (list): Inputs a evaluar en el conjunto de dato
            outputs (list): Outputs a evaluar en el conjunto de datos
            data (pandas.DataFrame): Conjunto de datos a evaluar
            C (float): Valores del hiperparámetro C del modelo
            eps (float): Valores del hiperparámetro épsilon del modelo
            d (int): Valor del hiperparámetro d del modelo
        """
        super().__init__(method, inputs, outputs, data, C, eps, d)

    def train(self):
        """Metodo que entrena un modelo SVF Splines
        """

        y_df = self.data.filter(self.outputs)
        y = y_df.values.tolist()

        # Numero de dimensiones y del problema
        n_out = len(y_df.columns)
        # Numero de observaciones del problema
        n_obs = len(y)
        n_inp = len(self.inputs)

        #######################################################################
        # Crear el grid
        self.grid = SVFSplinesGrid(self.data, self.inputs, self.outputs, self.d)
        self.grid.create_grid()

        # Número de variables w
        n_var = 0
        for i in range(len(self.grid.data_grid["phi"][0][0])):
            var_dim = self.grid.data_grid["phi"][0][0][i]
            n_var += len(var_dim)
        # Variable w
        name_w = [(knot, inp, out) for out in range(n_out) for inp in range(n_inp) for knot in
                  range(len(self.grid.knot_list[inp]) + 1)]

        w = {}
        w = w.fromkeys(name_w, 1)
        # Variable Xi
        name_xi = [(obs, out) for obs in range(n_obs) for out in range(n_out)]
        xi = {}
        xi = xi.fromkeys(name_xi, self.C)

        mdl = Model("SVF Splines")
        mdl.context.cplex_parameters.threads = 1

        # Variable w
        w_var = mdl.continuous_var_dict(name_w, ub=1e+33, lb=-1e+33, name='w')
        # Variable xi
        xi_var = mdl.continuous_var_dict(name_xi, ub=1e+33, lb=0, name='xi')

        # Función objetivo
        mdl.minimize(mdl.sum((w_var[i] * w[i]) ** 2 for i in name_w) + mdl.sum(xi_var[i] * xi[i] for i in name_xi))

        # Restricciones
        for obs in range(n_obs):
            for out in range(n_out):
                left_side = -y[obs][out] + mdl.sum(
                    w_var[knot, inp, out] * self.grid.data_grid.phi[obs][out][inp][knot] for inp in range(n_inp) for
                    knot in range(len(self.grid.knot_list[inp]) + 1))
                # (1)
                mdl.add_constraint(
                    left_side <= self.eps + xi_var[obs, out],
                    ctname='c1_o' + str(obs + 1) + "_y" + str(out + 1)
                )
                # (2)
                mdl.add_constraint(
                    -left_side <= 0,
                    ctname='c2_o' + str(obs + 1) + "_y" + str(out + 1)
                )

        for out in range(n_out):
            #     left_side = w_var[0, 0, 0]
            #     mdl.add_constraint(
            #         left_side >= 0,
            #         ctname='c3_x' + str(1) + '_y' + str(out + 1)
            #     )
            for inp in range(n_inp):
                for knot in range(2, len(self.grid.knot_list[inp]) + 2):
                    left_side = mdl.sum(w_var[knot, inp, out] * 1 for knot in range(1, knot))
                    # (3)
                    mdl.add_constraint(
                        left_side >= 0,
                        ctname='c3_x' + str(inp + 1) + '_y' + str(out + 1)
                    )

        for out in range(n_out):
            for inp in range(n_inp):
                # for i in range(len(t[inp]) + 1):
                for knot in range(1, len(self.grid.knot_list[inp]) + 1):
                    # (4)
                    if knot <= 1:
                        mdl.add_constraint(
                            w_var[knot, inp, out] >= 0,
                            'c4_x' + str(inp + 1) + "_y" + str(out + 1)
                        )
                    else:
                        mdl.add_constraint(
                            w_var[knot, inp, out] <= 0,
                            'c4_x' + str(inp + 1) + "_y" + str(out + 1)
                        )

        self.model = mdl
        if self.model_d is None:
            self.model_d = mdl

    def modify_model(self, c, eps):
        """Método que se utiliza para modificar el valor de C y las restricciones de un modelo
        Args:
            c (float): Valores del hiperparámetro C del modelo_
            eps (float): Valores del hiperparámetro épsilon del modelo

        Returns:
            docplex.mp.model.Model: modelo SVF modificado
        """

        n_out = len(self.outputs)
        n_dmu = len(self.data)
        model = self.model_d.copy()
        model.name = "SVF,C:" + str(c) + ",eps:" + str(eps) + ",d:" + str(self.d)
        name_var = model.iter_variables()
        name_w = list()
        name_xi = list()
        for var in name_var:
            name = var.get_name()
            if name.find("w") == -1:
                name_xi.append(name)
            else:
                name_w.append(name)
        # Variable w
        w = {}
        w = w.fromkeys(name_w, 1)
        # Variable Xi
        xi = {}
        xi = xi.fromkeys(name_xi, c)
        a = [(model.get_var_by_name(i) * w[i]) ** 2 for i in name_w]

        b = [model.get_var_by_name(i) * xi[i] for i in name_xi]
        # Función objetivo
        model.minimize(model.sum(a) + model.sum(b))
        # Modificar restricciones
        for out in range(n_out):
            for dmu in range(n_dmu):
                const_name = 'c1_o' + str(dmu + 1) + "_y" + str(out + 1)
                rest = model.get_constraint_by_name(const_name)
                rest.rhs += eps
        return model

    def solve(self):
        """Solución de un modelo SVF
        """
        n_out = len(self.outputs)
        self.model.solve()
        name_var = self.model.iter_variables()
        sol_w = list()
        sol_xi = list()
        for var in name_var:
            name = var.get_name()
            sol = self.model.solution[name]
            if name.find("w") == -1:
                sol_xi.append(sol)
            else:
                sol_w.append(sol)
        mat_w = [[] for _ in range(n_out)]
        cont = 0
        for i in range(n_out):
            for j in range(len(self.grid.data_grid["phi"][i][0][0])):
                w = round(sol_w[cont], 6)
                mat_w[i].append(w)
                cont += 1

        mat_xi = [[] for _ in range(n_out)]
        cont = 0
        for i in range(n_out):
            for j in range(len(self.data)):
                mat_xi[i].append(round(sol_xi[cont], 6))
                cont += 1

        self.solution = SVFPrimalSolution(mat_w, sol_xi)

    def get_estimation(self, dmu):
        """Estimacion de una DMU escogida. y=phi(x)*w
        Args:
            dmu (list): Observación sobre la que estimar su valor
        Returns:
            list: Devuelve una lista con la estimación de cada output
        """
        if len(dmu) != len(self.inputs):
            raise RuntimeError("El número de inputs de la DMU no coincide con el número de inputs del problema.")
        phi = self.grid.calculate_dmu_phi(dmu)
        prediction_list = list()
        for out in range(len(self.outputs)):
            # print(self.solution.w[out], phi[out])
            prediction = round(sum([a * b for a, b in zip(self.solution.w[out], phi[out][0])]), 3)
            prediction_list.append(prediction)
        return prediction_list





from svf_package.svf_functions import create_SVF
from tests import data


if __name__ == '__main__':

    inputs = [["x1"],["x1","x2"],["x1","x2","y1"]]
    outputs = [["y1"]]
    prediccion = [[3],[1,2],[1,3,1]]
    c = 1
    eps = 0
    d = 2
    method_list = ["SVF","SSVF","dual","SVF-SP"]
    for i in inputs:
        cont = len(i)-1
        for o in outputs:
            for method in method_list:
                print("N_INP:", i," N_OUT:", o)
                svf = create_SVF(method, i, o, data, c, eps, d)
                svf.train()
                # print(svf.model.export_to_string())
                svf.solve()
                print(method, svf.solution.w, "=>", svf.get_estimation(prediccion[cont]))
                print("***********************************************************")
    print("=======================================================================")




class SVFPrimalSolution:
    """Clase solución de SVF
    """
    def __init__(self, w, xi):
        """Constructor de la clase solución SVF

        Args:
            w (list): Valores de los pesos w del modelo ya solucionado
            xi (list): Valores xi del modelo ya solucionado
        """
        self.w = w
        self.xi = xi



from docplex.mp.model import Model
from numpy import dot, array, around
from svf_package.grid.svfgrid import SVFGrid
from svf_package.methods.svf import SVF
from svf_package.solution.svf_dual_solution import SVFDualSolution


class SVFDual(SVF):

    def __init__(self, method, inputs, outputs, data, C, eps, d):
        """Constructor de la clase SVFDual

        Args:
            method (string): Método SVF que se quiere utilizar
            inputs (list): Inputs a evaluar en el conjunto de dato
            outputs (list): Outputs a evaluar en el conjunto de datos
            data (pandas.DataFrame): Conjunto de datos a evaluar
            C (float): Valores del hiperparámetro C del modelo
            eps (float): Valores del hiperparámetro épsilon del modelo
            d (int): Valor del hiperparámetro d del modelo
        """
        super().__init__(method, inputs, outputs, data, C, eps, d)

    def train(self):
        """Método que entrena un modelo SVF en su forma dual
        """

        outputs_df = self.data.filter(self.outputs)
        y = outputs_df.values.tolist()

        # Numero de dimensiones y del problema
        n_out = len(outputs_df.columns)
        # Numero de observaciones del problema
        n_obs = len(y)

        #######################################################################
        # Matriz de t y de indices de ts
        # Crear el grid
        self.grid = SVFGrid(self.data, self.inputs, self.outputs, self.d)
        self.grid.create_grid()

        # Numero de variables w
        n_var = len(self.grid.data_grid.phi[0][0])

        #######################################################################

        # Variable alpha
        name_alpha = [(obs, out) for obs in range(0, n_obs) for out in range(0, n_out)]
        alpha = {}
        alpha = alpha.fromkeys(name_alpha, 1)

        # Variable delta
        name_delta = [(obs, out) for obs in range(0, n_obs) for out in range(0, n_out)]
        delta = {}
        delta = delta.fromkeys(name_delta, 1)

        # Variable gamma
        name_gamma = [(var, out) for var in range(0, n_var) for out in range(0, n_out)]

        mdl = Model("SVF DUAL C:" + str(self.C) + ", eps:" + str(self.eps) + ", d:" + str(self.d))
        mdl.context.cplex_parameters.threads = 1

        # Variable alpha
        alpha_var = mdl.continuous_var_dict(name_alpha, ub=1e+33, lb=0, name='alpha')
        # Variable delta
        delta_var = mdl.continuous_var_dict(name_delta, ub=self.C, lb=0, name='delta')
        # Variable gamma
        gamma_var = mdl.continuous_var_dict(name_gamma, ub=1e+33, lb=0, name='gamma')

        # Funcion objetivo
        s1 = mdl.sum(
            (alpha_var[obs, out] * alpha[obs, out] - delta_var[obs, out] * delta[obs, out]) *
            (alpha_var[obs2, out] * alpha[obs2, out] - delta_var[obs2, out] * delta[obs2, out]) *
            (dot(self.grid.data_grid.phi[obs][out], self.grid.data_grid.phi[obs2][out])) for out in range(n_out)
            for obs in range(n_obs) for obs2 in range(n_obs)
        )
        s2 = mdl.sum(
            2 * (alpha_var[obs, out] * alpha[obs, out] - delta_var[obs, out] * delta[obs, out]) *
            mdl.sum(self.grid.data_grid.phi[obs][out][var] * gamma_var[var, out])
            for out in range(n_out) for obs in range(n_obs) for var in range(n_var)
        )
        s3 = mdl.sum(gamma_var[var, out] * gamma_var[var, out] for out in range(n_out) for var in range(n_var))
        s4 = self.eps * mdl.sum(delta_var[obs, out] * delta[obs, out] for out in range(n_out) for obs in range(n_obs))
        s5 = mdl.sum(
            (alpha_var[obs, out] * alpha[obs, out] - delta_var[obs, out] * delta[obs, out]) * y[obs][out] for out in
            range(n_out) for obs in
            range(n_obs))

        mdl.minimize(1 / 2 * (s1 + s2 + s3) + s4 - s5)

        self.model = mdl

    def solve(self):
        """Método que soluciona el modelo entrenado.
        """
        n_obs = len(self.data)
        n_var = len(self.grid.data_grid.phi[0][0])
        n_out = len(self.outputs)
        self.model.solve()

        name_var = self.model.iter_variables()
        sol_gamma = list()
        sol_alpha = list()
        sol_delta = list()
        for var in name_var:
            name = var.get_name()
            sol = self.model.solution[name]
            if name.find("gamma") == 0:
                sol_gamma.append(sol)
            elif name.find("alpha") == 0:
                sol_alpha.append(sol)
            elif name.find("delta") == 0:
                sol_delta.append(sol)

        mat_sol_gamma = [[] for _ in range(0, n_out)]
        mat_sol_alpha = [[] for _ in range(0, n_out)]
        mat_sol_delta = [[] for _ in range(0, n_out)]

        cont = 0
        cont2 = 0
        for out in range(n_out):
            for var in range(n_var):
                mat_sol_gamma[out].append(round(sol_gamma[cont], 6))
                cont += 1
            for k in range(n_obs):
                mat_sol_alpha[out].append(round(sol_alpha[cont2], 6))
                mat_sol_delta[out].append(round(sol_delta[cont2], 6))
                cont2 += 1

        sumatorio = 0
        mat_sol_w = [[] for _ in range(0, n_out)]
        for out in range(n_out):
            for obs in range(n_obs):
                sumatorio += (mat_sol_alpha[out][obs] - mat_sol_delta[out][obs]) * array(
                    self.grid.data_grid.phi[obs][out])
            sol = sumatorio + array(mat_sol_gamma[out])
            mat_sol_w[out] = around(sol, 3).tolist()

        self.solution = SVFDualSolution(mat_sol_gamma, mat_sol_alpha, mat_sol_delta, mat_sol_w)



from docplex.mp.model import Model
from numpy import asarray, float32
from svf_package.grid.svfgrid import SVFGrid
from svf_package.methods.svf import SVF
from svf_package.solution.svf_solution import SVFPrimalSolution


class SVFC(SVF):
    """Clase del modelo SVF Splines
    """

    def __init__(self, method, inputs, outputs, data, C, eps, d):
        """Constructor de la clase SVF Splines

        Args:
            method (string): Método SVF que se quiere utilizar
            inputs (list): Inputs a evaluar en el conjunto de dato
            outputs (list): Outputs a evaluar en el conjunto de datos
            data (pandas.DataFrame): Conjunto de datos a evaluar
            C (float): Valores del hiperparámetro C del modelo
            eps (float): Valores del hiperparámetro épsilon del modelo
            d (int): Valor del hiperparámetro d del modelo
        """
        super().__init__(method, inputs, outputs, data, C, eps, d)

    def train(self):

        y_df = self.data.filter(self.outputs)
        y = y_df.values.tolist()

        # Numero de dimensiones y del problema
        n_out = len(y_df.columns)
        # Numero de observaciones del problema
        n_obs = len(y)

        # Crear el grid
        self.grid = SVFGrid(self.data, self.inputs, self.outputs, self.d)
        self.grid.create_grid()
        # Numero de variables w
        n_var = len(self.grid.data_grid.phi[0][0])

        # Variable u
        name_u = [(out, var) for out in range(n_out) for var in range(n_var)]
        u = {}
        u = u.fromkeys(name_u, 1)
        # Variable u
        name_v = [(out, var) for out in range(n_out) for var in range(n_var)]
        v = {}
        v = v.fromkeys(name_v, 1)
        # Variable Xi
        name_xi = [(out, var) for out in range(n_out) for var in range(n_obs)]
        xi = {}
        xi = xi.fromkeys(name_xi, self.C)
        mdl = Model("SVF C:" + str(self.C) + ", eps:" + str(self.eps) + ", d:" + str(self.d))
        mdl.context.cplex_parameters.threads = 1

        # Variable w
        u_var = mdl.continuous_var_dict(name_u, ub=1e+33, lb=0, name='u')
        v_var = mdl.continuous_var_dict(name_u, ub=1e+33, lb=0, name='v')
        # Variable xi
        xi_var = mdl.continuous_var_dict(name_xi, ub=1e+33, lb=0, name='xi')

        # Funcion objetivo
        mdl.minimize(mdl.sum(u_var[i] * u[i] for i in name_u) +
                     mdl.sum(v_var[i] * v[i] for i in name_v) +
                     mdl.sum(xi_var[i] * xi[i] for i in name_xi))
        # Restricciones
        for obs in range(n_obs):
            for out in range(n_out):
                left_side = y[obs][out] - \
                            mdl.sum(
                                u_var[out, var] * self.grid.data_grid.phi[obs][out][var] -
                                v_var[out, var] * self.grid.data_grid.phi[obs][out][var] for var in range(n_var)
                            )
                # (1)
                mdl.add_constraint(
                    left_side <= 0,
                    ctname='c1_' + str(obs) + "_" + str(out)
                )
                # (2)
                mdl.add_constraint(
                    -left_side <= self.eps + xi_var[out, obs],
                    ctname='c2_' + str(obs) + "_" + str(out)
                )
        # (3)
        for out in range(n_out):
            lhs = mdl.sum(
                u_var[out, var] * self.grid.df_grid.phi[0][out][var] - v_var[out, var] * self.grid.df_grid.phi[0][out][
                    var] for var in
                range(n_var)
            )
            mdl.add_constraint(
                lhs >= 0,
                ctname='c3_' + str(self.grid.df_grid.id_cell[0]) + "_" + str(out)
            )
        for index, cell in self.grid.df_grid.iterrows():
            left_side = cell["phi"][0]
            c_cont = cell["c_cells"]
            for c_cell in c_cont:
                c_cont_row = self.grid.df_grid.loc[self.grid.df_grid['id_cell'] == c_cell]
                right_side = c_cont_row["phi"].values[0][0]
                constraint = asarray(left_side, dtype=float32) - asarray(right_side, dtype=float32)
                for out in range(n_out):
                    lhs = mdl.sum(
                        u_var[out, var] * constraint[var] - v_var[out, var] * constraint[var] for var in range(n_var)
                    )
                    mdl.add_constraint(
                        lhs >= 0,
                        ctname='c3_' + str(cell["id_cell"]) + "_" + str(out)
                    )
        self.model = mdl
        if self.model_d is None:
            self.model_d = mdl

    def solve(self):
        """Solución de un modelo SVF
        """
        n_out = len(self.outputs)
        self.model.solve()
        name_var = self.model.iter_variables()
        sol_u = list()
        sol_v = list()
        sol_xi = list()
        for var in name_var:
            name = var.get_name()
            sol = self.model.solution[name]
            if name.find("u") == 0:
                sol_u.append(sol)
            elif name.find("v") == 0:
                sol_v.append(sol)
            else:
                sol_xi.append(sol)
        # Numero de ws por dimension
        n_w_dim = int(len(sol_u) / n_out)
        mat_w = [[] for _ in range(n_out)]
        cont = 0
        for out in range(n_out):
            for w_dim in range(n_w_dim):
                w = sol_u[cont] - sol_v[cont]
                mat_w[out].append(w)
                cont += 1
        mat_xi = [[] for _ in range(n_out)]
        cont = 0
        for out in range(n_out):
            for obs in range(len(self.data)):
                mat_xi[out].append(round(sol_xi[cont], 6))
                cont += 1
        self.solution = SVFPrimalSolution(mat_w, mat_xi)



class SVF:
    """Clase del algoritmo Support Vector Frontiers
    """

    def __init__(self, method, inputs, outputs, data, C, eps, d):
        """Constructor de la clase SVF
        Args:
            method (string): Método SVF que se quiere utilizar
            inputs (list): Inputs a evaluar en el conjunto de dato
            outputs (list): Outputs a evaluar en el conjunto de datos
            data (pandas.DataFrame): Conjunto de datos a evaluar
            C (float): Valores del hiperparámetro C del modelo
            eps (float): Valores del hiperparámetro épsilon del modelo
            d (int): Valor del hiperparámetro d del modelo
        """
        self.method = method
        self.data = data
        self.outputs = outputs
        self.inputs = inputs
        self.C = C
        self.eps = eps
        self.d = d
        self.grid = None
        self.model = None
        self.model_d = None
        self.solution = None
        self.name = None

    def modify_model(self, c, eps):
        """Método que se utiliza para modificar el valor de C y las restricciones de un modelo
        Args:
            c (float): Valores del hiperparámetro C del modelo_
            eps (float): Valores del hiperparámetro épsilon del modelo

        Returns:
            docplex.mp.model.Model: modelo SVF modificado
        """

        n_obs = len(self.data)
        model = self.model_d.copy()
        model.name = "SVF,C:" + str(c) + ",eps:" + str(eps) + ",d:" + str(self.d)
        name_var = model.iter_variables()
        name_w = list()
        name_xi = list()
        for var in name_var:
            name = var.get_name()
            if name.find("xi") == 0:
                name_xi.append(name)
            else:
                name_w.append(name)
        # Variable w
        w = {}
        w = w.fromkeys(name_w, 1)
        # Variable Xi
        xi = {}
        xi = xi.fromkeys(name_xi, c)

        a = [model.get_var_by_name(i) * w[i] for i in name_w]
        b = [model.get_var_by_name(i) * xi[i] for i in name_xi]
        # Funcion objetivo
        model.minimize(model.sum(a) + model.sum(b))
        # Modificar restricciones
        for obs in range(n_obs):
            for r in range(len(self.outputs)):
                const_name = 'c2_' + str(obs) + "_" + str(r)
                rest = model.get_constraint_by_name(const_name)
                rest.rhs += eps
        return model

    def get_estimation(self, dmu):
        """Estimacion de una DMU escogida. y=phi(dmu)*w

        Args:
            dmu (list): Observación sobre la que estimar su valor

        Returns:
            list: Devuelve una lista con la estimación de cada output
        """
        if len(dmu) != len(self.inputs):
            raise RuntimeError("El número de inputs de la DMU no coincide con el número de inputs del problema.")
        dmu_cell = self.grid.search_dmu(dmu)
        phi = self.grid.df_grid.loc[self.grid.df_grid['id_cell'] == dmu_cell, "phi"].values[0]
        prediction_list = list()
        for out in range(len(self.outputs)):
            prediction = round(sum([a * b for a, b in zip(self.solution.w[out], phi[out])]), 3)
            prediction_list.append(prediction)
        return prediction_list



from tests import data
from svf_package.grid.svf_splines_grid import SVFSplinesGrid

if __name__ == '__main__':

    inputs = ["x1","x2"]
    outputs = ["y1"]
    d = 2

    grid = SVFSplinesGrid(data,inputs,outputs,d)
    grid.create_grid()

    print(grid.data_grid)


from pathlib import Path
from setuptools import setup, find_packages

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

VERSION = '0.0.4'
DESCRIPTION = 'Paquete Support Vector Frontier'
PACKAGE_NAME = 'svf_package'
AUTHOR = 'Daniel Valero Carreras'
EMAIL = 'dvalero@umh.es'
GITHUB_URL = 'https://github.com/danielvacarre'

setup(
    name = PACKAGE_NAME,
    packages = find_packages(),
    version = VERSION,
    license='MIT',
    description = DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    author = AUTHOR,
    author_email = EMAIL,
    url = GITHUB_URL,
    keywords = [],
    install_requires=[
        "docplex", "joblib","numpy","pandas","python-dateutil","pytz",
        "scikit-learn", "scipy", "six", "threadpoolctl"
    ],
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Mathematics',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
)


from svf_package.methods.ssvf import SSVF
from svf_package.methods.svf_dual import SVFDual
from svf_package.methods.svf_splines import SVFSplines
from svf_package.methods.svfc import SVFC


def calculate_mse(svf, data_test):
    """Función que calcula el Mean Square Error (MSE) del cross-validation

        Args:
            data_test (pandas.DataFrame): conjunto de datos de test sobre los que se va a evaluar el MSE
            svf (svf_package.svf.SVF): modelo SVF sobre el que se va a evaluar los datos de test. Contiene los pesos (w) y el grid para calcular la estimación

        Returns:
            float: Mean Square Error obtenido para ese modelo y conjunto de datos
        """
    data_test_X = data_test.filter(svf.inputs)
    data_test_Y = data_test.filter(svf.outputs)
    n_out = len(data_test_Y.columns)
    error = 0
    n_obs_test = len(data_test_X)
    for i in range(n_obs_test):
        dmu = data_test_X.iloc[i]
        y_est = svf.get_estimation(dmu)
        for j in range(n_out):
            y = data_test_Y.iloc[i, j]
            error_obs = (y - y_est[j]) ** 2
            error = error + error_obs
    mse = error / n_obs_test
    return mse


def create_SVF(method, inputs, outputs, data, c, eps, d):
    """Función que crea un objeto del tipo SVF en función del método que se selecciona
    Args:
        method (string): Método SVF que se quiere utilizar
        inputs (list): Inputs a evaluar en el conjunto de dato
        outputs (list): Outputs a evaluar en el conjunto de datos
        data (pandas.DataFrame): Conjunto de datos a evaluar
        c (float): Valores del hiperparámetro C del modelo
        eps (float): Valores del hiperparámetro épsilon del modelo
        d (int): Valor del hiperparámetro d del modelo

    Raises:
        RuntimeError: Indica que no existe el método seleccionado

    Returns:
        object: Devuelve un objeto del método SVF seleccionado
    """
    if method == "SVF-SP":
        svf = SVFSplines(method, inputs, outputs, data, c, eps, d)
    elif method == "SSVF":
        svf = SSVF(method, inputs, outputs, data, c, eps, d)
    elif method == "SVF":
        svf = SVFC(method, inputs, outputs, data, c, eps, d)
    elif method == "dual":
        svf = SVFDual(method, inputs, outputs, data, c, eps, d)
    else:
        raise RuntimeError("The method selected doesn't exist")
    return svf



from docplex.mp.model import Model
from svf_package.grid.svfgrid import SVFGrid
from svf_package.methods.svf import SVF
from svf_package.solution.svf_solution import SVFPrimalSolution


class SSVF(SVF):
    """Clase del modelo SVF Simplificado
    """

    def __init__(self, method, inputs, outputs, data, C, eps, d):
        """Constructor de la clase SSVF

        Args:
            method (string): Método SVF que se quiere utilizar
            inputs (list): Inputs a evaluar en el conjunto de dato
            outputs (list): Outputs a evaluar en el conjunto de datos
            data (pandas.DataFrame): Conjunto de datos a evaluar
            C (float): Valores del hiperparámetro C del modelo
            eps (float): Valores del hiperparámetro épsilon del modelo
            d (int): Valor del hiperparámetro d del modelo
        """
        super().__init__(method, inputs, outputs, data, C, eps, d)

    def train(self):
        """Metodo que entrena un modelo SSVF
        """

        y_df = self.data.filter(self.outputs)
        y = y_df.values.tolist()

        # Numero de dimensiones y del problema
        n_out = len(y_df.columns)
        # Numero de observaciones del problema
        n_obs = len(y)

        #######################################################################
        # Crear el grid
        self.grid = SVFGrid(self.data, self.inputs, self.outputs, self.d)
        self.grid.create_grid()

        # Numero de variables w
        n_var = len(self.grid.data_grid.phi[0][0])

        #######################################################################

        # Variable w
        # name_w: (i,j)-> i:es el indice de la columna de la matriz phi;j: es el indice de la dimension de y
        name_w = [(out, w_var) for out in range(n_out) for w_var in range(n_var)]
        w = {}
        w = w.fromkeys(name_w, 1)

        # Variable Xi
        name_xi = [(out, obs) for out in range(n_out) for obs in range(n_obs)]
        xi = {}
        xi = xi.fromkeys(name_xi, self.C)

        mdl = Model("SSVF C:" + str(self.C) + ", eps:" + str(self.eps) + ", d:" + str(self.d))
        mdl.context.cplex_parameters.threads = 1

        # Variable w
        w_var = mdl.continuous_var_dict(name_w, ub=1e+33, lb=0, name='w')
        # Variable xi
        xi_var = mdl.continuous_var_dict(name_xi, ub=1e+33, lb=0, name='xi')

        # Funcion objetivo
        mdl.minimize(mdl.sum(w_var[i] * w[i] for i in name_w) + mdl.sum(xi_var[i] * xi[i] for i in name_xi))

        # Restricciones
        for obs in range(n_obs):
            for out in range(n_out):
                left_side = y[obs][out] - mdl.sum(w_var[out, var] * self.grid.data_grid.phi[obs][out][var]
                                                  for var in range(n_var))
                # (1)
                mdl.add_constraint(
                    left_side <= 0,
                    ctname='c1_' + str(obs) + "_" + str(out)
                )
                # (2)
                mdl.add_constraint(
                    -left_side <= self.eps + xi_var[out, obs],
                    ctname='c2_' + str(obs) + "_" + str(out)
                )
        self.model = mdl
        if self.model_d is None:
            self.model_d = mdl

    def solve(self):
        """Solución de un modelo SVF
        """

        n_out = len(self.outputs)
        self.model.solve()
        name_var = self.model.iter_variables()
        sol_w = list()
        sol_xi = list()
        for var in name_var:
            name = var.get_name()
            sol = self.model.solution[name]
            if name.find("w") == -1:
                sol_xi.append(sol)
            else:
                sol_w.append(sol)
        # Numero de ws por dimension
        n_w_dim = int(len(sol_w) / n_out)
        mat_w = [[] for _ in range(n_out)]
        cont = 0
        for out in range(n_out):
            for j in range(0, n_w_dim):
                mat_w[out].append(round(sol_w[cont], 6))
                cont += 1
        mat_xi = [[] for _ in range(n_out)]
        cont = 0
        for out in range(n_out):
            for j in range(0, len(self.data)):
                mat_xi[out].append(round(sol_xi[cont], 6))
                cont += 1
        self.solution = SVFPrimalSolution(mat_w, mat_xi)
